// Copyright E. Kowalski 2010-2015
// License: GPLv2, see file COPYING in the source tar file

package net.frogparrot.kloostermania;

import java.math.BigInteger;

import static net.frogparrot.kloostermania.KloosView.nextPrime;

/**
* Created by emmanuel on 4/24/15.
*/
class KloosData {

    Double[] kr, ki, sr, si; // Kloosterman real and imaginary parts, and same for Birch

    // We now use only a single argument
    int m,c;
    int sumType;
    double[] bds;


    public KloosData(int m, int c, int s) {
        double xmin,xmax,ymin,ymax;

        this.sumType = s;

        //Log.v("KL",c + " Create");
        this.m = m;
        //        this.n = n;

        // We enforce the primality of the modulus
        this.c = nextPrime(c);

        this.kr=new Double[c];
        this.ki=new Double[c];
        kr[0]=0.0;
        ki[0]=0.0;
        if (sumType != Kloostermania.KLOOS_TYPE_KLOOS) {
            // Birch sums require c+1 points since we also store the origin
            this.sr=new Double[c+1];
            this.si=new Double[c+1];
            sr[0]=0.0;
            si[0]=0.0;
        }

        int[] j;
        xmin = -0.1;
        xmax = 0.1;
        ymin = -0.1;
        ymax = 0.1;

        double dx = 1.0, dy = 1.0;
        int icube;

        // Birch sum; we must be careful with overflow
        // This is rather painful...
        if (sumType != Kloostermania.KLOOS_TYPE_KLOOS) {
            BigInteger bc = BigInteger.valueOf(c);
            BigInteger three = BigInteger.valueOf(3);
            for(int i=1; i<c+1; i++) {
                BigInteger ic = BigInteger.valueOf(i-1);
                ic = ic.modPow(three,bc);
                icube = ic.intValue();
                sr[i] = (double) (sr[i-1]+Math.cos(2*Math.PI/c*(m*(i-1)+icube)));
                if (sr[i]<xmin) { xmin = sr[i]; }
                if (sr[i]>xmax) { xmax = sr[i]; }

                si[i] = (double) (si[i-1]+Math.sin(2*Math.PI/c*(m*(i-1)+icube)));
                if (si[i]<ymin) { ymin = si[i]; }
                if (si[i]>ymax) { ymax = si[i]; }
            }
        }
        dx = xmax-xmin;
        dy = ymax-ymin;


        // Start again if needed
        if (sumType != Kloostermania.KLOOS_TYPE_BIRCH) {
            for(int i=1; i<c; i++) {
                j = KloosView.egcd(i, c);
                // We keep the test although c should be prime...
                if (j[0] == 1) {
                    kr[i] = (double) (kr[i-1]+Math.cos(2*Math.PI/c*(m*i+j[1])));
                    if (kr[i]<xmin) { xmin = kr[i]; }
                    if (kr[i]>xmax) { xmax = kr[i]; }

                    ki[i] = (double) (ki[i-1]+Math.sin(2*Math.PI/c*(m*i+j[1])));
                    if (ki[i]<ymin) { ymin = ki[i]; }
                    if (ki[i]>ymax) { ymax = ki[i]; }
                }
            }
            dx = xmax-xmin;
            dy = ymax-ymin;
        }

        // Expand a bit the size on both sides for cleaner display
        bds = new double[] {xmin-0.1*dx, xmax+0.1*dx, ymin-0.15*dy, ymax+0.15*dy};

    }

    // Constructor without Birch sum, just copy the Kloosterman case

    public KloosData(int m, int c) {
        double xmin,xmax,ymin,ymax;

        sumType = Kloostermania.KLOOS_TYPE_KLOOS;
        //Log.v("KL",c + " Create");
        this.m = m;
        this.c = c;

        this.kr=new Double[c];
        this.ki=new Double[c];

        int[] j;
        xmin = -0.1;
        xmax = 0.1;
        ymin = -0.1;
        ymax = 0.1;
        kr[0]=0.0;
        ki[0]=0.0;

        for(int i=1; i<c; i++) {
            j = KloosView.egcd(i, c);
            if (j[0] == 1) {
                kr[i] = (double) (kr[i-1]+Math.cos(2*Math.PI/c*(m*i+j[1])));
                if (kr[i]<xmin) { xmin = kr[i]; }
                if (kr[i]>xmax) { xmax = kr[i]; }

                ki[i] = (double) (ki[i-1]+Math.sin(2*Math.PI/c*(m*i+j[1])));
                if (ki[i]<ymin) { ymin = ki[i]; }
                if (ki[i]>ymax) { ymax = ki[i]; }
            }
        }
        double dx = xmax-xmin;
        double dy = ymax-ymin;
        // Expand a bit the size on both sides for cleaner display
        bds = new double[] {xmin-0.1*dx, xmax+0.1*dx, ymin-0.15*dy, ymax+0.15*dy};
    }
}
