// Copyright E. Kowalski 2010-2015
// License: GPLv2, see file COPYING in the source tar file

package net.frogparrot.kloostermania;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.view.View;

/**
* Created by emmanuel on 4/24/15.
*/


class KloosView extends View {

    //Context ctx;
    KloosData k;
    //RandomFourierSeries sample;
    Paint rfs;
    Paint axes;
    Paint curvek;
    Paint curves;
    // True if we draw axes
    boolean drawAxes = true;

    // Some mathematical utilities

    public static int[] egcd(int p, int q) {
        if (q == 0)
            return new int[] { p, 1, 0 };

        int[] vals = egcd(q, p % q);
        int d = vals[0];
        int a = vals[2];
        int b = vals[1] - (p / q) * vals[2];
        return new int[] { d, a, b };
    }

    public static boolean isPrime(int p) {
        int j, k;

        if (p==2) { return true; }
        if (p%2==0) { return false; }
        k = (int) Math.sqrt(p)+1;
        for(j=3;j<=k;j+=2) {
            if (p%j==0) { return false; }
        }
        return true;
    }

    public static int nextPrime(int p) {
    	/* Finds smallest prime >= p */
        int n=p;
        if (n<0) { n=2; }
        if (n<=2) { return 2; }
        while (!isPrime(n)) {
            n=n+1;
        }
        return n;
    }

    public static int previousPrime(int p) {
    	/* Finds largest prime <=p */
        int n=p;
        if (n<0) { n=2; }
        if (n<=3) { return 2; }
        while (!isPrime(n)) {
            n=n-1;
        }
        return n;
    }

    public KloosView(Context context) {
        super(context);

        // Color for the axes
        axes = new Paint();
        axes.setColor(Color.GRAY);
        axes.setStrokeWidth(2);
        // Kloostermania is red
        curvek = new Paint();
        curvek.setColor(Color.RED);
        curvek.setStrokeWidth(2);
        curvek.setAntiAlias(true);
        curvek.setStrokeCap(Paint.Cap.ROUND);
        // Birch is Blue
        curves = new Paint();
        curves.setColor(Color.BLUE);
        curves.setStrokeWidth(2);
        curves.setAntiAlias(true);
        curves.setStrokeCap(Paint.Cap.ROUND);

        // Random Fourier series
        rfs = new Paint();
        rfs.setColor(Color.GREEN);
        rfs.setStrokeWidth(2);
        rfs.setAntiAlias(true);
        rfs.setStrokeCap(Paint.Cap.ROUND);

        setFocusable(true);
    }

    public void createKloos(int a, int q, int typesum) {

        //Log.v("KL",q+ " create kloos");
        //if (onlyPrimes == true) { m = nextPrime(m); }
        k = new KloosData(a,q, typesum);
        // TODO: use the right length
        //sample = new RandomFourierSeries();
    }

    @Override
    public void onDraw(Canvas canvas) {

        //Log.v("KL","onDraw "+ k.c);
        canvas.drawColor(Color.WHITE);
        int h=canvas.getHeight();
        int w=canvas.getWidth();

        if (k != null) {
            double dx=k.bds[1]-k.bds[0];
            double dy=k.bds[3]-k.bds[2];
            float xsc = (float) dx/w;
            float ysc = (float) dy/h;

            if (xsc>ysc) {
                // Width largest: use horizontal pixel as basic unit
                float pixs = (float) dx/w;
                double ratio = h*pixs/dy;
                k.bds[2] = k.bds[2] * ratio;
                k.bds[3] = k.bds[3] * ratio;
            } else {
                // Height is largest
                float pixs = (float) dy / h;
                double ratio = w * pixs / dx;
                k.bds[1] = k.bds[1] * ratio;
                k.bds[0] = k.bds[0] * ratio;
            }

            // Recompute x and y pixel scales
            xsc = (float) (k.bds[1]-k.bds[0])/w;
            // These will now essentially be the same!
            ysc = (float) (k.bds[3]-k.bds[2])/h;

            float cx = (float) -k.bds[0]/xsc;
            float cy = (float) ((float) h+k.bds[2]/ysc);

            // Draw the axes
            if (drawAxes) {
                canvas.drawLine(cx, 0, cx, h, axes);
                canvas.drawLine(0, cy, w, cy, axes);

                // Draw ticks on x axis
                int normdx = (int) Math.round((w / (dx / Math.sqrt(k.c))));

                for (int i = 0; i < Math.round(k.bds[1]); i++) {
                    canvas.drawLine(cx + i * normdx, cy - 10, cx + i * normdx, cy + 10, axes);
                }

                for (int i = 0; i < Math.round(-k.bds[0]); i++) {
                    canvas.drawLine(cx - i * normdx, cy - 10, cx - i * normdx, cy + 10, axes);
                }


                // Draw ticks on y axis
                int normdy = (int) Math.round((h / (dy / Math.sqrt(k.c))));

                for (int i = 0; i < Math.round(k.bds[3]); i++) {
                    canvas.drawLine(cx - 10, cy + i * normdy, cx + 10, cy + i * normdy, axes);
                }

                for (int i = 0; i < Math.round(-k.bds[2]); i++) {
                    canvas.drawLine(cx - 10, cy - i * normdy, cx + 10, cy - i * normdy, axes);
                }
            }
            // Loop over the part of the arrays that actually exists...
            if (k.sumType != Kloostermania.KLOOS_TYPE_BIRCH) {
                for (int i=0; i<k.c-1 ; i++) {

                    float sx = cx+k.kr[i].floatValue()/xsc;
                    float sy = cy-k.ki[i].floatValue()/ysc;
                    float ex = cx+k.kr[i+1].floatValue()/xsc;
                    float ey = cy-k.ki[i+1].floatValue()/ysc;
                    canvas.drawLine(sx,sy,ex,ey,curvek);

                }
            }
            // Same for Birch
            if (k.sumType != Kloostermania.KLOOS_TYPE_KLOOS) {
                for (int i=0; i<k.c ; i++) {

                    float sx = cx+k.sr[i].floatValue()/xsc;
                    float sy = cy-k.si[i].floatValue()/ysc;
                    float ex = cx+k.sr[i+1].floatValue()/xsc;
                    float ey = cy-k.si[i+1].floatValue()/ysc;
                    canvas.drawLine(sx,sy,ex,ey,curves);

                }
            }

            // Draw sample

       /*     double step;
            step = 1.0/k.c;
            ComplexNumber val1, val2;
            val1 = new ComplexNumber(0.0,0.0);
            for (int i=0; i<k.c; i++){
                val2 = sample.evaluate((i + 1) * step);
                float sx = cx+(float) (val1.Real()*Math.sqrt(k.c))/xsc;
                float sy = cy-(float) (val1.Im()*Math.sqrt(k.c))/ysc;
                float ex = cx+(float) (val2.Real()*Math.sqrt(k.c))/xsc;
                float ey = cy-(float) (val2.Im()*Math.sqrt(k.c))/ysc;
                canvas.drawLine(sx,sy,ex,ey,rfs);
                val1=val2;
            }*/

        }
        invalidate();
    }


}
