// Copyright E. Kowalski 2010-2015
// License: GPLv2, see file COPYING in the source tar file

package net.frogparrot.kloostermania;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.ActivityNotFoundException;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.res.AssetManager;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.drawable.Drawable;
import android.net.Uri;
import android.os.Handler;
import android.os.Message;
import android.os.Bundle;
import android.text.Html;
import android.text.method.LinkMovementMethod;
import android.view.GestureDetector;
import android.view.Menu;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.View;
import android.widget.ShareActionProvider;
import android.widget.TextView;
import android.widget.Toast;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Random;

import static net.frogparrot.kloostermania.KloosView.previousPrime;
import static net.frogparrot.kloostermania.KloosView.nextPrime;


public class Kloostermania extends Activity {

    KloosView kv;
    // Version number
    private static final String version = "1.01";
    // A random number generator
    private static Random rng;
   // True is slideshow is playing
    boolean slideShow = false;


    // Constants for the options meanu
    private static final int MENU_NEW_KLOOS = 0;
    private static final int MENU_ABOUT = 1;
    private static final int MENU_MOD = 2;
    private static final int MENU_RANDOM = 4;
    private static final int MENU_PRIMES = 5;
    private static final int MENU_SLIDE = 6;
    private static final int MENU_SHARE = 7;
    private static final int MENU_ZEN = 8;

    // Constants for the type of sums shown
    public final static int KLOOS_TYPE_KLOOS = 0;
    public final static int KLOOS_TYPE_BIRCH = 1;
    public final static int KLOOS_TYPE_BOTH = 2;
    // Default is only Kloosterman

    // Messages for slideshow
    public static final int KLOOS_MSG_STOP = 0;
    public static final int KLOOS_MSG_RUNNING = 1;
    public static final int KLOOS_ARG2_STOP = 0;
    public static final int KLOOS_ARG2_CONTINUE = 1729;
    // Delay for slideshow
    private static final int slideShowTime = 2500;

    // Constant for the Intent when inputing a new modulus
    private static final int CHOOSE_SUM = 2;

    private GestureDetector gesture;

    private ShareActionProvider sharer;


    // Slideshow handler
    private Handler slideshow = new Handler(new Handler.Callback() {

        public boolean handleMessage(Message msg) {
            switch (msg.what) {
                case KLOOS_MSG_STOP: // Stop slideshow -- currently unused!
                    slideShow = false;
                    return true;
                case KLOOS_MSG_RUNNING: // Slideshow is running
                    if (msg.arg2 == KLOOS_ARG2_CONTINUE) {
                        slideShow = true;
                        Message nmsg = Message.obtain();
                        nmsg.what = 1;
                        // Is it really needed?
                        nmsg.arg1 = kv.k.c+1;
                        nmsg.arg2 = KLOOS_ARG2_CONTINUE;
                        kv.createKloos(kv.k.m,nextPrime(kv.k.c+1),kv.k.sumType);
                        boolean unused = slideshow.sendMessageDelayed(nmsg, slideShowTime);
                        showModulusAndSum();
                        kv.invalidate();
                        // And continue...
                    } else {
                        slideShow = false;
                    }
                    return true;
            }
            return false;
        }
    });

    // Utility to display modulus and sum quickly

    public String infoSums() {

        String message = "";
        switch (kv.k.sumType) {
            case KLOOS_TYPE_KLOOS:
                message = "Kloosterman sum: ";
                break;
            case KLOOS_TYPE_BIRCH:
                message = "Birch sum: ";
                break;
            case KLOOS_TYPE_BOTH:
                break;
        }

        message = message + "Modulus = " +  Integer.toString(kv.k.c) + "; a= " +
                Integer.toString(kv.k.m)+ ";\n";

        switch (kv.k.sumType) {
            case KLOOS_TYPE_KLOOS:
                message = message + "Sum = " + kv.k.kr[kv.k.c-1]
                        + "\nNormal = " + kv.k.kr[kv.k.c-1]/Math.sqrt(kv.k.c);
                break;
            case KLOOS_TYPE_BIRCH:
                message  = message + "Sum = " + kv.k.sr[kv.k.c]
                        + "\nNormal = " + kv.k.sr[kv.k.c]/Math.sqrt(kv.k.c);
                break;
            case KLOOS_TYPE_BOTH:
                message = message + "K = " + kv.k.kr[kv.k.c-1] + "  (red)"
                        + "\nNormal = " + kv.k.kr[kv.k.c-1]/Math.sqrt(kv.k.c)
                        + "\nB = " + kv.k.sr[kv.k.c] + "  (blue)"
                        + "\nNormal = " + kv.k.sr[kv.k.c]/Math.sqrt(kv.k.c) ;
                break;
        }
        return message;
    }


    public void showModulusAndSum () {
        String message = infoSums();
        Toast toast = Toast.makeText(kv.getContext(),message,Toast.LENGTH_SHORT);
        toast.show();
    }

    // Callback class for scroll and tap
    private class KloosGestureListener extends GestureDetector.SimpleOnGestureListener {

        float MINIMAL = 100;
        private float x1,x2;
        private float y1,y2;

        @Override
        public boolean onSingleTapConfirmed(MotionEvent event) {
            int screenw, screenh, edgel,edger,edgeb,edget;

            // Single press
            //Log.d("onSingleTapUp",event.toString());

            // If we tap close to an edge, we implement scroll
            int[] loc = new int[2];
            kv.getLocationOnScreen(loc);

            screenw = kv.getWidth();
            screenh = kv.getHeight();
            edgel = screenw / 10;
            edger = 9*screenw / 10;
            edget = loc[1]+screenh /10;
            edgeb = loc[1]+9*screenh/10;

            x1=event.getX();
            y1=event.getY();

            if (x1<edgel) {
                int c = previousPrime(kv.k.c - 1);
                kv.createKloos(kv.k.m, c, kv.k.sumType);
                kv.invalidate();
                showModulusAndSum();
                return true;
            }

            if (x1>edger) {
                int c = nextPrime(kv.k.c + 1);
                kv.createKloos(kv.k.m, c, kv.k.sumType);
                kv.invalidate();
                showModulusAndSum();
                return true;
            }

            if (y1>edgeb) {
                int a = (kv.k.m-1)%kv.k.c;
                kv.createKloos(a,kv.k.c,kv.k.sumType);
                kv.invalidate();
                showModulusAndSum();
                return true;
            }

            if (y1<edget){
                int a = (kv.k.m+1)%kv.k.c;
                kv.createKloos(a,kv.k.c,kv.k.sumType);
                kv.invalidate();
                showModulusAndSum();
                return true;
            }

            // Now look for middle tap

            switch (event.getAction()) {
                case MotionEvent.ACTION_UP:
                    // For singleTap the action is DOWN.
                case MotionEvent.ACTION_DOWN:
                    if (slideShow) {
                        Message msg = Message.obtain();
                        msg.what = KLOOS_MSG_RUNNING;
                        msg.arg1 = 2; // Should be modulus, but not used since we stop...
                        // Stop the show
                        msg.arg2 = KLOOS_ARG2_STOP;
                        slideshow.removeMessages(1);
                        slideshow.sendMessage(msg);
                        // Indicate the show is stopped...
                        Toast toast = Toast.makeText(kv.getContext(), "Slideshow stopped.",Toast.LENGTH_SHORT);
                        toast.show();
                        return true;
                    } else {
                        showModulusAndSum();
                        return true;
                    }
            }
            return true;
        }

        @Override
        public boolean onDoubleTap(MotionEvent e)
        {
            switch (kv.k.sumType) {
                case KLOOS_TYPE_BIRCH:
                    kv.createKloos(kv.k.m, kv.k.c, KLOOS_TYPE_BOTH);
                    break;
                case KLOOS_TYPE_KLOOS:
                    kv.createKloos(kv.k.m, kv.k.c, KLOOS_TYPE_BIRCH);
                    break;
                case KLOOS_TYPE_BOTH:
                    kv.createKloos(kv.k.m, kv.k.c, KLOOS_TYPE_KLOOS);
                    break;
            }
            kv.invalidate();
            return true;
        }

        @Override
        public boolean onScroll(MotionEvent e1, MotionEvent e2, float distanceX, float distanceY) {
            int c;
            float x1,x2,y1,y2,dx,dy;

            x1=e1.getX();
            x2=e2.getX();
            y1=e1.getY();
            y2=e2.getY();
            dx=Math.abs(x1-x2);
            dy=Math.abs(y1-y2);

            //Log.d("onScroll",Float.toString(distanceX));
            //Log.d("onScroll",Float.toString(distanceY));

            if (dx>MINIMAL) {
                if (x1 > x2) {
                    c = nextPrime(kv.k.c + 1);
                    kv.createKloos(kv.k.m, c, kv.k.sumType);
                    kv.invalidate();
                } else {
                    c = previousPrime(kv.k.c - 1);
                    kv.createKloos(kv.k.m, c, kv.k.sumType);
                    kv.invalidate();
                }
                return true;
            }

            // Vertical: change a
            if (dy>MINIMAL) {
                if (y2>y1) {
                    int a = (kv.k.m+1)%kv.k.c;
                    kv.createKloos(a,kv.k.c,kv.k.sumType);
                    kv.invalidate();
                } else {
                    int a = (kv.k.m-1)%kv.k.c;
                    kv.createKloos(a,kv.k.c,kv.k.sumType);
                    kv.invalidate();
                }
                return true;
            }


            return true;
        }

    }

    /** Called when the activity is first created. */
    @Override
    public void onCreate(Bundle savedInstanceState) {

        super.onCreate(savedInstanceState);
        gesture = new GestureDetector(this, new KloosGestureListener());
        rng = new Random();
        kv = new KloosView(this);

        // Restore settings
        SharedPreferences settings = getSharedPreferences("Kloostermania", 0);
        int modulus = settings.getInt("modulus", 173);
        int m = settings.getInt("m",1);
        int n = settings.getInt("n",1);
        // Old version saved the onlyPrimes setting so we check that the modulus is...
        modulus = nextPrime(modulus);
        slideShow = settings.getBoolean("slide", false);

        int typesum = KLOOS_TYPE_KLOOS;
        try {
            // Older versions used a boolean setting for Birch sums
            typesum = settings.getInt("sumType",KLOOS_TYPE_KLOOS);
        } catch (ClassCastException e) {
            boolean s = settings.getBoolean("sumType", false);
            if (s) {
                typesum = KLOOS_TYPE_BIRCH;
            } else {
                typesum = KLOOS_TYPE_KLOOS;
            }
        }
        // Old versions did not have a toggle for axes drawing
        try {
            kv.drawAxes = settings.getBoolean("axes",false);
        } catch (Exception e) {
            kv.drawAxes = true;
        }

        // Create the view and display
        kv.createKloos(m,modulus,typesum);

        // Don't start earlier since the data is not there yet otherwise...
        if (slideShow) {
            startSlideShow();
        }
        setContentView(kv);
    }

    @Override
    public boolean onTouchEvent(MotionEvent event) {
        return gesture.onTouchEvent(event);
    }

    public void startSlideShow() {

        // Clear up the queue, if any
        slideshow.removeMessages(1);
        Message msg = Message.obtain();
        msg.what = KLOOS_MSG_RUNNING;
        msg.arg1 = kv.k.c;
        msg.arg2 = KLOOS_ARG2_CONTINUE;
        // Start sliding in 2.5 seconds
        showModulusAndSum();
        boolean unused = slideshow.sendMessageDelayed (msg, slideShowTime);

    }

    // onPause is where the orientation change works...
    @Override
    protected void onPause(){
        super.onPause();

        // Stop the slideshow if running...
        if (slideShow) {
            slideshow.removeMessages(1);
        }
        // Save the settings
        SharedPreferences settings = getSharedPreferences("Kloostermania", 0);
        SharedPreferences.Editor editor = settings.edit();
        editor.putInt("modulus", kv.k.c);
        editor.putBoolean("slide", slideShow);
        editor.putInt("sumType", kv.k.sumType);
        editor.putInt("m", kv.k.m);
        //editor.putInt("n", kv.k.n);
        editor.putBoolean("axes",kv.drawAxes);
        // Commit must not be forgotten
        editor.commit();
    }


    // Listen for results.
    protected void onActivityResult(int requestCode, int resultCode, Intent data){

        // Get the modulus
        int modulus = data.getIntExtra("net.frogparrot.kloostermania.modulus", 179);
        modulus = nextPrime(modulus);
        int m = data.getIntExtra("net.frogparrot.kloostermania.m",1);
        int typesum = data.getIntExtra("net.frogparrot.kloostermania.type", KLOOS_TYPE_KLOOS);
        kv.createKloos(m,modulus,typesum);
        kv.invalidate();
    }

    /* Creates the menu items; onPrepare... is called every time the menu is opened, so
     * it is better for the dynamic toggle of the moduli... */

    public boolean onPrepareOptionsMenu(Menu menu) {

        menu.clear();

        menu.add(0, MENU_NEW_KLOOS, 0, "New sum");
        menu.add(0,MENU_SHARE,0,"Share");
        if (kv.drawAxes) {
            menu.add(0, MENU_ZEN, 0, "Zen mode");
        } else {
            menu.add(0,MENU_ZEN,0,"Show axes");
        }
        // Group 1 will help disable the menu during slideshow
        menu.add(1, MENU_SLIDE, 0, "Slideshow");
        menu.add(0, MENU_RANDOM, 0, "Random");

        menu.add(0, MENU_MOD, 0, "Modulus and sum");
        menu.add(0, MENU_ABOUT, 0, "About");
        // Disable during slideshow
        if (slideShow) { menu.setGroupEnabled(1,false); }
        return true;
    }

    public void shareSum () {

        //Define a bitmap with the same size as the view
        Bitmap sumdraw = Bitmap.createBitmap(kv.getWidth(), kv.getHeight(),Bitmap.Config.ARGB_8888);
        //Bind a canvas to it
        Canvas canvas = new Canvas(sumdraw);
        //Get the view's background
        Drawable bgDrawable =kv.getBackground();
        if (bgDrawable!=null)
            //has background drawable, then draw it on the canvas
            bgDrawable.draw(canvas);
        else
            //does not have background drawable, then draw white background on the canvas
            canvas.drawColor(Color.WHITE);
        // draw the view on the canvas
        kv.draw(canvas);

        File tempFile = new File(getExternalFilesDir(null), "temp.png");
        try {
            FileOutputStream out = new FileOutputStream(tempFile);
            sumdraw.compress(Bitmap.CompressFormat.PNG, 100, out);
            out.flush();
            out.close();
        }
        catch (Exception e) {
        }

        Intent shareIntent = new Intent();
        shareIntent.setAction(Intent.ACTION_SEND);
        shareIntent.putExtra(Intent.EXTRA_STREAM,Uri.fromFile(tempFile));
        shareIntent.setType("image/png");
        shareIntent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION);
        startActivity(Intent.createChooser(shareIntent, "send"));

    }

    public void showPDFPaper () {
        File pdfFile = new File(getExternalFilesDir(null), "k.pdf");
        Intent pdfIntent;

        try {
            // Very simple code to copy a picture from the application's
            // resource into the external pdfFile. Note that if external storage is
            // not currently mounted this will silently fail.
            AssetManager am = getAssets();
            InputStream is = am.open("k.pdf");
            OutputStream os = new FileOutputStream(pdfFile);
            byte[] data = new byte[is.available()];
            int nbread=is.read(data);
            os.write(data);
            is.close();
            os.close();
            // Now we start an intent to show the PDF pdfFile...
            Uri path = Uri.fromFile(pdfFile);
            pdfIntent = new Intent(Intent.ACTION_VIEW);
            pdfIntent.setDataAndType(path, "application/pdf");
            pdfIntent.setFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP);
            startActivity(pdfIntent);
        } catch (IOException e) {
            // Unable to create pdfFile, likely because external storage is
            // not currently mounted.
        } catch (ActivityNotFoundException e) {
            Toast.makeText(kv.getContext(), "No Application available to view pdf",
                    Toast.LENGTH_LONG).show();
        }
    }

    /* Handles item selections */
    public boolean onOptionsItemSelected(MenuItem item) {
        int modulus = 173;

        switch (item.getItemId()) {

            case MENU_NEW_KLOOS:
                Intent intent = new Intent(this, net.frogparrot.kloostermania.Choosesum.class);
                intent.putExtra("net.frogparrot.kloostermania.modulus", kv.k.c);
                intent.putExtra("net.frogparrot.kloostermania.type",kv.k.sumType);
                intent.putExtra("net.frogparrot.kloostermania.m",kv.k.m);
                //intent.putExtra("net.frogparrot.kloostermania.n",kv.k.n);
                startActivityForResult(intent, CHOOSE_SUM);
                return true;
            case MENU_SLIDE:
                slideShow = true;
                //invalidateOptionsMenu();
                startSlideShow();
                return true;
            case MENU_RANDOM:
                // Biased dice to select smaller moduli most of the time...
                int choice = rng.nextInt(10);
                switch (choice) {
                    case 0:
                        modulus = 10000+rng.nextInt(10000);
                        break;
                    case 1:
                    case 2:
                    case 3:
                        modulus = 5000 + rng.nextInt(5000);
                        break;
                    case 4:
                    case 5:
                    case 6:
                        modulus = 1000+rng.nextInt(4000);
                    case 7:
                    case 8:
                    case 9:
                        modulus = 2+rng.nextInt(1000);
                }
                modulus = nextPrime(modulus);
                kv.createKloos(kv.k.m,modulus,kv.k.sumType);
                showModulusAndSum();
                kv.invalidate();
                return true;
            case MENU_MOD:
                new AlertDialog.Builder(this)
                        .setMessage(infoSums())
                        .setPositiveButton("OK", new DialogInterface.OnClickListener() {
                            public void onClick(DialogInterface dialog) {
                            }
                            public void onClick(DialogInterface dialog, int which) {
                            }})
                        .setOnCancelListener(new DialogInterface.OnCancelListener() {
                            public void onCancel(DialogInterface dialog) {
                            }})
                        .show();
                return true;
            case MENU_ABOUT:
                View view = View.inflate(this, R.layout.about, null);
                TextView textView = (TextView) view.findViewById(R.id.abouttext);

                textView.setMovementMethod(LinkMovementMethod.getInstance());
                textView.setText(Html.fromHtml("<b>Pocket Kloostermania</b><br>Version " +
                                version +
                                "<br>Paper: <i>Kloosterman paths and the shape of exponential sums</i>, E. Kowalski and W. Sawin" +
                                "<br>License: GPLv2" +
                                "<br>See <a href=\"http://blogs.ethz.ch/kowalski/the-kloostermania-page\">blogs.ethz.ch/kowalski/the-kloostermania-page</a>"));

                new AlertDialog.Builder(this)
                        .setView(view)
                        .setPositiveButton("OK", new DialogInterface.OnClickListener() {
                            public void onClick(DialogInterface dialog) {
                            }

                            public void onClick(DialogInterface dialog, int which) {
                            }
                        })
                        .setNeutralButton("Open PDF", new DialogInterface.OnClickListener() {
                            public void onClick(DialogInterface dialog) {
                                showPDFPaper();
                            }

                            public void onClick(DialogInterface dialog, int which) {
                                showPDFPaper();
                            }
                        })
                        .setOnCancelListener(new DialogInterface.OnCancelListener() {
                            public void onCancel(DialogInterface dialog) {
                            }})
                        .show();
                return true;
            case MENU_SHARE:
                shareSum();
                return true;
            case MENU_ZEN:
                kv.drawAxes=!kv.drawAxes;
                return true;
        } // Switch

        //We do not need a settings item
        //return super.onOptionsItemSelected(item);
        return false;
    }

}

