// Copyright E. Kowalski 2010-2015
// License: GPLv2, see file COPYING in the source tar file

package net.frogparrot.kloostermania;

import java.util.Random;

public class RandomFourierSeries {

    Double[] coefficients;
    int length;
    boolean isInitialized = false;
    Random rng;

    double[] bds;

    public RandomFourierSeries () {

        initialize(100);
    }

    public RandomFourierSeries (int l) {

        initialize(l);

    }


    public void initialize (int l) {

        rng = new Random();

        if (l>=1) { length=l; } else { length = 100; }

        coefficients = new Double[2*length+1];

        for (int i=0; i<2*length+1;i++) {
            coefficients[i] = 2 * Math.cos(2 * Math.PI * rng.nextDouble());
        }

        isInitialized = true;
    }

    public ComplexNumber evaluate(double t) {

        // We must sum
        // 1/(2iPi h) * (e(ht)-1) * (random coeffs)

        double aux=2*Math.PI*t;
        ComplexNumber value = new ComplexNumber(0.0, 0.0);
        double c = 1/2.0/Math.PI;

        if (t==0.0) { return value; }
        if (isInitialized) {
            double real = 0.0;
            double imag = 0.0;


            for (int i =-length;i<=length; i++) {
                if (i==0) {
                    real = real + t*coefficients[length];
                }
                else {
                    real = real + c/i * Math.sin(aux*i) * coefficients[i+length];
                    imag = imag - c/i * (Math.cos(aux*i)-1) * coefficients[i+length];
                }
            }
            value.setReal(real);
            value.setIm(imag);
        }

        return value;

    }

}
